open Bitstring_utils
open ELF_types
open ELF_printers
open ELF_utils
open Library
open PPC_parsers

exception Unknown_endianness

(** Converts an elf endian into a bitstring endian *)
let elfdata_to_endian (e: elfdata): Bitstring.endian =
  match e with
  | ELFDATA2LSB -> Bitstring.LittleEndian
  | ELFDATA2MSB -> Bitstring.BigEndian
  | _           -> raise Unknown_endianness

(** Parses an elf32 header *)
let read_elf32_ehdr (bs: bitstring): elf32_ehdr =
  bitmatch bs with
  | { e_ident : 16*8 : bitstring ;
      rest    : -1   : bitstring } ->
      (
        bitmatch e_ident with
        | { 0x7F       : 8              ;
            "ELF"      : 24 : string    ;
            ei_class   : 8  : int       ;
            ei_data    : 8  : int       ;
            ei_version : 8  : int       ;
            padding    : 72 : bitstring } ->
            assert (is_zeros padding 72);
            let ei_data =
              begin match ei_data with
              | 0 -> ELFDATANONE
              | 1 -> ELFDATA2LSB
              | 2 -> ELFDATA2MSB
              | _ -> ELFDATAUNKNOWN
              end
            in
            let e = elfdata_to_endian ei_data in
            (
              bitmatch rest with
              | { e_type      : 16 : int, endian(e) ;
                  e_machine   : 16 : int, endian(e) ;
                  e_version   : 32 : int, endian(e) ;
                  e_entry     : 32 : int, endian(e) ;
                  e_phoff     : 32 : int, endian(e) ;
                  e_shoff     : 32 : int, endian(e) ;
                  e_flags     : 32 : bitstring      ;
                  e_ehsize    : 16 : int, endian(e) ;
                  e_phentsize : 16 : int, endian(e) ;
                  e_phnum     : 16 : int, endian(e) ;
                  e_shentsize : 16 : int, endian(e) ;
                  e_shnum     : 16 : int, endian(e) ;
                  e_shstrndx  : 16 : int, endian(e) } ->
                  (* These shouldn't be different than this... *)
                  assert (e_ehsize = 52);
                  assert (e_phentsize = 32);
                  assert (e_shentsize = 40);
                  {
                    e_ident =
                      {
                        ei_class      =
                          begin match ei_class with
                          | 0 -> ELFCLASSNONE
                          | 1 -> ELFCLASS32
                          | 2 -> ELFCLASS64
                          | _ -> ELFCLASSUNKNOWN
                          end;
                        ei_data     = ei_data;
                        ei_version  =
                          begin match ei_version with
                          | 0 -> EV_NONE
                          | 1 -> EV_CURRENT
                          | _ -> EV_UNKNOWN
                          end;
                      };
                    e_type        =
                      begin match e_type with
                      | 0 -> ET_NONE
                      | 1 -> ET_REL
                      | 2 -> ET_EXEC
                      | 3 -> ET_DYN
                      | 4 -> ET_CORE
                      | _ -> ET_UNKNOWN
                      end;
                    e_machine     =
                      begin match e_machine with
                      | 0  -> EM_NONE
                      | 1  -> EM_M32
                      | 2  -> EM_SPARC
                      | 3  -> EM_386
                      | 4  -> EM_68K
                      | 5  -> EM_88K
                      | 7  -> EM_860
                      | 8  -> EM_MIPS
                      | 10 -> EM_MIPS_RS4_BE
                      | 20 -> EM_PPC
                      | _  -> EM_UNKNOWN
                      end;
                    e_version     =
                      begin match e_version with
                      | 0l -> EV_NONE
                      | 1l -> EV_CURRENT
                      | _ -> EV_UNKNOWN
                      end;
                    e_entry     = e_entry;
                    e_phoff     = e_phoff;
                    e_shoff     = e_shoff;
                    e_flags     = e_flags;
                    e_ehsize    = e_ehsize;
                    e_phentsize = e_phentsize;
                    e_phnum     = e_phnum;
                    e_shentsize = e_shentsize;
                    e_shnum     = e_shnum;
                    e_shstrndx  = e_shstrndx;
                  }
            )
      )

(** Returns the file offset of the section header indexed *)
let section_header_offset (e_hdr: elf32_ehdr) (sndx: int): int =
  Safe.(of_int32 e_hdr.e_shoff + (sndx * e_hdr.e_shentsize))

(** Returns the ndx-th string in the provided bitstring, according to null
    characters *)
let strtab_string (bs: bitstring) (ndx: int): string =
  let (str, ofs, _) = bs in
  let start = (ofs / 8 + ndx) in
  String.sub str start (String.index_from str start '\000' - start)

(** Reads an ELF section header *)
let read_elf32_shdr (e_hdr: elf32_ehdr) (bs: bitstring) (strtab: bitstring)
    (num: int): elf32_shdr =
  let e = elfdata_to_endian e_hdr.e_ident.ei_data in
  let bit_ofs = Safe.(
    (section_header_offset e_hdr num) * 8
  ) in
  bitmatch bs with
  | { sh_name      : 32 : endian(e), offset(bit_ofs) ;
      sh_type      : 32 : endian(e) ;
      sh_flags     : 32 : endian(e) ;
      sh_addr      : 32 : endian(e) ;
      sh_offset    : 32 : endian(e) ;
      sh_size      : 32 : endian(e) ;
      sh_link      : 32 : endian(e) ;
      sh_info      : 32 : endian(e) ;
      sh_addralign : 32 : endian(e) ;
      sh_entsize   : 32 : endian(e) } ->
      {
        sh_name         = strtab_string strtab (Safe32.to_int sh_name);
        sh_type         =
          begin match sh_type with
          |  0l -> SHT_NULL
          |  1l -> SHT_PROGBITS
          |  2l -> SHT_SYMTAB
          |  3l -> SHT_STRTAB
          |  4l -> SHT_RELA
          |  5l -> SHT_HASH
          |  6l -> SHT_DYNAMIC
          |  7l -> SHT_NOTE
          |  8l -> SHT_NOBITS
          |  9l -> SHT_REL
          | 10l -> SHT_SHLIB
          | 11l -> SHT_DYNSYM
          | _   -> SHT_UNKNOWN
          end;
        sh_flags     = sh_flags     ;
        sh_addr      = sh_addr      ;
        sh_offset    = sh_offset    ;
        sh_size      = sh_size      ;
        sh_link      = sh_link      ;
        sh_info      = sh_info      ;
        sh_addralign = sh_addralign ;
        sh_entsize   = sh_entsize   ;
      }

let string_of_instruction = function
| Padd      (i0, i1, i2)        -> "Padd(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Padde     (i0, i1, i2)        -> "Padde(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Paddi     (i0, i1, c2)        -> "Paddi(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_constant c2 ^ ")"
| Paddic    (i0, i1, c2)        -> "Paddic(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_constant c2 ^ ")"
| Paddis    (i0, i1, c2)        -> "Paddis(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_constant c2 ^ ")"
| Paddze    (i0, i1)            -> "Paddze(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ")"
| Pallocframe(c0, i1)            -> "Pallocframe(" ^ string_of_coq_Z c0 ^ ", " ^ string_of_iint i1 ^ ")"
| Pand_     (i0, i1, i2)        -> "Pand_(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Pandc     (i0, i1, i2)        -> "Pandc(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Pandi_    (i0, i1, c2)        -> "Pandi_(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_constant c2 ^ ")"
| Pandis_   (i0, i1, c2)        -> "Pandis_(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_constant c2 ^ ")"
| Pb        (l0)                -> "Pb(" ^ string_of_label l0 ^ ")"
| Pbctr                         -> "Pbctr"
| Pbctrl                        -> "Pbctrl"
| Pbf       (c0, l1)            -> "Pbf(" ^ string_of_crbit c0 ^ ", " ^ string_of_label l1 ^ ")"
| Pbl       (i0)                -> "Pbl(" ^ string_of_ident i0 ^ ")"
| Pbs       (i0)                -> "Pbs(" ^ string_of_ident i0 ^ ")"
| Pblr                          -> "Pblr"
| Pbt       (c0, l1)            -> "Pbt(" ^ string_of_crbit c0 ^ ", " ^ string_of_label l1 ^ ")"
| Pbtbl     (i0, l1)            -> "Pbtbl(" ^ string_of_ireg i0 ^ ", " ^ string_of_list string_of_label ", " l1 ^ ")"
| Pcmplw    (i0, i1)            -> "Pcmplw(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ")"
| Pcmplwi   (i0, c1)            -> "Pcmplwi(" ^ string_of_ireg i0 ^ ", " ^ string_of_constant c1 ^ ")"
| Pcmpw     (i0, i1)            -> "Pcmpw(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ")"
| Pcmpwi    (i0, c1)            -> "Pcmpwi(" ^ string_of_ireg i0 ^ ", " ^ string_of_constant c1 ^ ")"
| Pcror     (c0, c1, c2)        -> "Pcror(" ^ string_of_crbit c0 ^ ", " ^ string_of_crbit c1 ^ ", " ^ string_of_crbit c2 ^ ")"
| Pdivw     (i0, i1, i2)        -> "Pdivw(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Pdivwu    (i0, i1, i2)        -> "Pdivwu(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Peqv      (i0, i1, i2)        -> "Peqv(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Pextsb    (i0, i1)            -> "Pextsb(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ")"
| Pextsh    (i0, i1)            -> "Pextsh(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ")"
| Pfreeframe(c0, i1)            -> "Pfreeframe(" ^ string_of_coq_Z c0 ^ ", " ^ string_of_iint i1 ^ ")"
| Pfabs     (f0, f1)            -> "Pfabs(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ")"
| Pfadd     (f0, f1, f2)        -> "Pfadd(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ", " ^ string_of_freg f2 ^ ")"
| Pfcmpu    (f0, f1)            -> "Pfcmpu(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ")"
| Pfcti     (i0, f1)            -> "Pfcti(" ^ string_of_ireg i0 ^ ", " ^ string_of_freg f1 ^ ")"
| Pfdiv     (f0, f1, f2)        -> "Pfdiv(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ", " ^ string_of_freg f2 ^ ")"
| Pfmake    (f0, i1, i2)        -> "Pfmake(" ^ string_of_freg f0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Pfmr      (f0, f1)            -> "Pfmr(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ")"
| Pfmul     (f0, f1, f2)        -> "Pfmul(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ", " ^ string_of_freg f2 ^ ")"
| Pfneg     (f0, f1)            -> "Pfneg(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ")"
| Pfrsp     (f0, f1)            -> "Pfrsp(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ")"
| Pfsub     (f0, f1, f2)        -> "Pfsub(" ^ string_of_freg f0 ^ ", " ^ string_of_freg f1 ^ ", " ^ string_of_freg f2 ^ ")"
| Plbz      (i0, c1, i2)        -> "Plbz(" ^ string_of_ireg i0 ^ ", " ^ string_of_constant c1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Plbzx     (i0, i1, i2)        -> "Plbzx(" ^ string_of_ireg i0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Plfd      (f0, c1, i2)        -> "Plfd(" ^ string_of_freg f0 ^ ", " ^ string_of_constant c1 ^ ", " ^ string_of_ireg i2 ^ ")"
| Plfdx     (f0, i1, i2)        -> "Plfdx(" ^ string_of_freg f0 ^ ", " ^ string_of_ireg i1 ^ ", " ^ string_of_ireg i2 ^ ")"